'''
Run fasttree.

Functions:
    run_fasttree(filename, outfile=None, fasttree_location='fasttree') -> None

'''
from getphylo.utils import io

def run_fasttree(filename, outfile=None, fasttree_location='fasttree') -> None:
    '''
    Run fasttree on a protein alignment.
        Arguments:
            filename: path to the alignment
            outfile: path to the output file
        Returns:
            None
    '''

    if outfile is None:
        out = io.change_extension(filename, "tree")
    else:
        out = outfile
    command = [
        fasttree_location,
        "-out", out,
        filename
    ]
    io.run_in_command_line(command)



'''
Make DIAMOND databases and run searches using DIAMOND

Functions:
    make_diamond_database(filename: str, dmnd_database=None) -> None
    run_diamond_search(
    filename: str, dmnd_database=None, outname=None, diamond_args=['diamond',None,None,None]
    ) -> None
'''
import logging
from getphylo.utils import io

def make_diamond_database(infile: str, dmnd_database=None, diamond_location='diamond') -> None:
    '''
    Create a DIAMOND database from a fasta file.
        Arguments:
            filename: path to the input file
            dmnd_database: path to output directory
        Returns:
            None
    '''
    if dmnd_database is None:
        database_name = infile.split('.')[0] + ".dmnd"
    else:
        database_name = dmnd_database
    command = [
        diamond_location, "makedb",
        "--db", database_name, 
        "--in", infile
        ]
    logging.debug(command)
    io.run_in_command_line(command)

def run_diamond_search(
    filename: str, dmnd_database=None, outname=None, diamond_args=['diamond',None,None,None]
    ) -> None:
    '''
    Run BLASTP through DIAMOND.
        Arguments:
            filename: path to the input fasta file
            dmnd_database: name of a database file to read
            outname: name of the output file
            diamond_args: list of arguments for diamond
        Returns:
            None
    '''
    if dmnd_database is None:
        database = filename.split('.')[0] + ".dmnd"
    else:
        database = dmnd_database
    if outname is None:
        output = dmnd_database.split('.')[0] + ".tsv"
    else:
        output = outname
    command = [
        diamond_args[0], "blastp",
        "--db", database,
        "--query", filename,
        "--out", output,
        "--outfmt", "6", "qseqid", "sseqid", "pident"
        ]
    #there must be a nicer way of doing this but it works for now!
    if diamond_args[1] is not None:
        command.append("--id")
        command.append(str(diamond_args[1]))
    if diamond_args[2] is not None:
        command.append("--query-cover")
        command.append(str(diamond_args[2]))
    if diamond_args[3] is not None:
        command.append("--subject-cover")
        command.append(str(diamond_args[3]))
    logging.debug(command)
    io.run_in_command_line(command)



'''
Input and output operations for getphylo.

Functions:
    get_locus(file:str, locus:str) -> str
    count_files(directory: str) -> int
    change_extension(filename: str, new_extension: str) -> str
    get_records_from_genbank(filename: str) -> List
    make_folder(name: str) -> None
    read_file(filename: str) -> List[str]
    read_tsv(filename: str) -> List[str]
    run_in_command_line(List[str])
    run_in_parallel(function: Callable, args_list: Iterable[List], cpus: int) -> List
    write_to_file(filename: str, write_lines: List[str]) -> None
'''
import csv
import glob
import multiprocessing
import os
import subprocess
import logging
from typing import Callable, Iterable, List
from Bio import SeqIO

from getphylo.utils.errors import GetphyloError, FolderExistsError, BadExecutableError

def get_locus(fasta: List[str], locus: str) -> str:
    '''
    Returns a sequence from a fasta file with the provided locus name.
        Arguments:
            fasta: list of lines from in fasta format
            locus: locus to search for
        Returns:
            sequence: the sequence of the locus
    '''
    line_number = 0
    sequence = ''
    for line in fasta:
        line_number += 1
        if locus in line:
            sequence = fasta[line_number]
            break
    assert sequence
    return sequence

def count_files(directory: str) -> int:
    '''
    Counts the number of files in a directory.
        Arguments:
            directory: the path to the directory in which to count
        Returns:
            number_of_files: the number of files in that directory as an integer
    '''
    number_of_files = len(glob.glob(directory))
    return number_of_files

def change_extension(filename: str, new_extension: str) -> str:
    '''
    Takes a file name and changes the extension to the string provided.
        Arguments:
            filename: the name of the file with the old extension
            new_extension: the new extension to replace with
        Returns:
            new_filename: the filename with the new extension
    '''
    new_filename = os.path.splitext(filename)[0] + '.' + new_extension
    return new_filename

def get_records_from_genbank(filename: str) -> List:
    '''
    Use BioPython to get genbank records from a given file.
        Arguments:
            filename: the filename of the genbank file
        Returns:
            records: a list genbank records
    '''
    records = SeqIO.parse(filename, "genbank")
    return records

def make_folder(name: str) -> None:
    '''
    Attempts to make a folder with the given name but raises and exception if it already exists.
        Arguments:
            name: the name of the folder being created
        Returns:
            None
    '''
    if os.path.exists(name):
        raise FolderExistsError(
            f'The directory {name} already exists.'
            'For saftey, please remove the folder before continuing.'
            )
    os.mkdir(name)

def read_file(filename: str) -> List[str]:
    '''
    Return a files contents as a list of lines
        Arguments:
            filename: The file to be read
        Returns:
            _file.readlines(): list of strings for each line of the file
    '''
    with open(filename) as _file:
        return _file.readlines()

def read_tsv(filename: str) -> List[str]:
    '''
    Read lines from a .tsv file.
        Arguments:
            filename: the path to the file to be read
        Returns:
            list containing the lines of the .tsv file
    '''
    contents = []
    with open(filename) as file:
        tsv_file = csv.reader(file, delimiter="\t")
        for line in tsv_file:
            contents.append(line)
    return contents

def run_in_command_line(command: List[str]) -> None:
    '''
    Convert a string into a command and run in the terminal.
        Aruments:
            command: list of strings containing the command for the terminal
        Returns:
            process: the process being run
    '''
    logging.debug(command)
    try:
        with subprocess.Popen(
            command, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE
            ) as process:
            _, stderr =process.communicate()
            if process.returncode != 0:
                raise RuntimeError(
                    'Failed to run: ' + str(command)
                    + 'with the following error ' + str(stderr))
            return process
    except FileNotFoundError as error:
        raise BadExecutableError(
            'getphylo could not find an executable, ' +
            'please ensure the correct paths to all executables are provided'
            ) from error

def run_in_parallel(function: Callable, args_list: Iterable[List], cpus: int) -> List:
    '''
    Run a given function on avaliable cpus. If only 1 cpu is available, run as normal.
        Arguments:
            function: the function to be called
            args_list: Iterable of lists containing the arguments for each call of the function
            cpus: the number of cpus available
        Returns:
            return_value: a list of return values for each call of the function
    '''
    if cpus <= 1:
        return_value = [function(*args) for args in args_list]
    else:
        for item in args_list:
            try:
                iter(item)
            except TypeError as error:
                raise GetphyloError from error
        try:
            with multiprocessing.Pool(cpus) as pool:
                results = pool.starmap_async(function, args_list)
                return_value = results.get()
        except Exception:
            raise
    return return_value

def write_to_file(filename: str, write_lines: List[str]) -> None:
    '''
    Write a list line by line into a new file.
        Arguments:
            filename: path to the new file being written
            write_lines: list of strings to be written to the file
        Returns:
            None
    '''
    with open(filename, "a") as _file:
        for line in write_lines:
            _file.write(line + '\n')
        _file.close()



'''
Create alignments from a folder of fasta files.

Functions:
    get_locus_from_tsv(locus: str, fasta_name: str) -> Tuple[str, str]
    make_fasta_for_alignments(loci_list: List[str], output:str) -> None
    do_alignments(output:str) -> None
    get_locus_length(alignment: List[str]) -> int
    get_locus_alignment(alignment, taxon_name) -> str
    make_combined_alignment(gbks: List, output: str) -> None
    make_alignments(checkpoint: Checkpoint, output: str, loci: List, gbks: List) -> None
'''
import logging
import os
import glob
from typing import List, Tuple
from getphylo.ext import muscle
from getphylo.utils import io
from getphylo.utils.checkpoint import Checkpoint
from getphylo.utils.errors import FileAlreadyExistsError, BadLocusError

def get_locus_from_tsv(locus: str, fasta_name: str) -> Tuple[str, str]:
    '''
    Extracts specific loci for the alignment.
        Arguments:
            locus: locus to extract
            fasta_name: the name of the .fasta file
        Returns:
            sequence_name: the locus and taxon name
            sequence: the sequence for aligning
    '''
    tsv_name = fasta_name.replace("/fasta/", "/tsvs/")
    tsv_name = io.change_extension(tsv_name, "tsv")
    tsv = io.read_tsv(tsv_name)
    fasta_contents = io.read_file(fasta_name)
    for line in tsv:
        if locus in line[0]:
            sequence = io.get_locus(fasta_contents, line[1])
            organism, _ = os.path.splitext(os.path.basename(fasta_name))
            sequence_name = f'>{organism}_{line[1]}'
            return sequence_name, sequence
    raise BadLocusError("Locus %s not found in file: %s" % (locus, tsv_name))

def make_fasta_for_alignments(loci_list: List[str], output: str) -> None:
    '''
    Builds a .fasta file from sequences where there is a hit in the diamond search results.
    Arguments:
        loci_list: a list of locus tags to extract hits
        output: the path of the output directory
    Returns:
        None
    '''
    io.make_folder(os.path.join(output, 'unaligned_fasta'))
    files = glob.glob(os.path.join(output, 'fasta/*.fasta'))
    assert files
    logging.debug(loci_list)
    for locus in loci_list:
        write_lines = []
        for filename in files:
            try:
                logging.debug('locus: %s; filename: %s' % (locus, filename))
                write_lines.extend(get_locus_from_tsv(locus, filename))
                logging.debug('write_lines %s', write_lines)
            except BadLocusError as error:
                logging.warning(error)

        outfile = os.path.join(output, 'unaligned_fasta', locus + '.fasta')
        io.write_to_file(outfile, write_lines)

def do_alignments(output: str, cpus: int, muscle_location: str) -> None:
    '''
    Runs the pre-aligned fasta file through the muscle module.
        Arguments:
            output: the path of the outut directory
        Returns:
            None
    '''
    io.make_folder(os.path.join(output, 'aligned_fasta'))
    args_list = []
    for filename in glob.glob(os.path.join(output, 'unaligned_fasta/*.fasta')):
        outfile = os.path.join(output, 'aligned_fasta', os.path.basename(filename))
        args_list.append([filename, outfile, muscle_location])
    io.run_in_parallel(muscle.run_muscle, args_list, cpus) #add cpu arg

def get_locus_length(alignment: List[str]) -> int:
    '''
    Returns the length of the fasta locus by counting sequence lines.
        Arguments:
            alignment: a list containing each line of a fasta file
        Returns:
            alignment_length: an int reprisenting the alignment length
    '''
    alignment_length = 0
    if not alignment:
        raise ValueError('An alignment cannot be empty.')
    for line in alignment[1:]:
        if '>' not in line:
            alignment_length += len(line.strip())
        else:
            break
    return alignment_length

def get_locus_alignment(alignment, taxon_name) -> str:
    '''
    Extract an alignment for a loci for a taxa.
        Arguments:
            alignment: List of lines representing the alignment fasta
            taxon_name: the name of the taxon to extract lines for
        Returns:
            extracted_alignment_string:
                the alignment of the specific loci and taxa combined as a string
    '''
    seq_flag = False
    extracted_alignment = []
    for line in alignment:
        if line[0] == '>':
            seq_flag = False
        if '>' in line and taxon_name in line:
            seq_flag = True
        if seq_flag and taxon_name not in line:
            extracted_alignment.extend(line[:-1])
    extracted_alignment_string = ''.join(extracted_alignment)
    return extracted_alignment_string

def format_partition_data(partition_data: List) -> List:
    '''
    Takes the partition data [locus, locus_length] and reformats it as a partition file 
    (e.g. p1 = 1-50, p2 = 51-110)
        Arguments:
            partition_data: a list of [locus, locus_length] lists
        Returns:
            partition_lines: list of lines in the partition format
    '''
    model = 'WAG'
    partition_lines = []
    partition_start = 1
    for partition in partition_data:
        locus, length = partition[0], partition[1]
        locus = os.path.splitext(os.path.basename(locus))[0]
        partition_end = partition_start + length - 1
        partition_lines.append('%s, %s = %s-%s' % (model, locus, partition_start, partition_end))
        partition_start += length
    return partition_lines

def make_combined_alignment(gbks: List, output: str) -> None:
    '''
    Create a combined alignment from single locus alignments
        Arguments:
            gbks: a list of genbank files
            output: path  to output directory
        Returns:
            None
    '''
    combined_alignment_path = os.path.join(output, 'aligned_fasta/combined_alignment.fasta')
    partition_path = os.path.join(output, 'partition.txt')
    if os.path.exists(combined_alignment_path):
        logging.error(
            'ALERT: aligned_fasta/combined_alignment.fasta already exists. Exiting!'
            )
        raise FileAlreadyExistsError('%s alread exists.' % combined_alignment_path)
    combined_alignment = []
    partition_data = []
    taxa = glob.glob(gbks)
    assert taxa, gbks
    loci = glob.glob(os.path.join(output, 'aligned_fasta/*.fasta'))
    first_taxa = True
    for taxon in taxa:
        taxon_name = os.path.splitext(os.path.basename(taxon))[0]
        sequence_data = []
        for locus in loci:
            alignment = io.read_file(locus)
            locus_length = get_locus_length(alignment)
            if first_taxa is True:
                partition_data.append([locus, locus_length])
            locus_alignment = get_locus_alignment(alignment, taxon_name)
            if len(locus_alignment) == locus_length:
                sequence_data.append(locus_alignment)
            else:
                sequence_data.append('?' * locus_length)
        sequence_string = ''.join(sequence_data)
        assert sequence_data
        if sequence_string.count('?') == len(sequence_string):
            logging.error('[ALERT]: %s has no sequence data and has been removed.', taxon_name)
        else:
            combined_alignment.append(f'>{taxon_name}')
            combined_alignment.append(sequence_string)
        first_taxa = False
    partition_data = format_partition_data(partition_data)
    io.write_to_file(combined_alignment_path, combined_alignment)
    io.write_to_file(partition_path, partition_data)

def make_alignments(
    checkpoint: Checkpoint, output: str, loci: List, gbks: List, cpus: int, muscle_location: str
    ) -> None:
    '''
    Main routine for align.
        Arguments:
            checkpoint: checkpoint defined by the user
            output: path to the output directory
            loci: list of loci to align
            gbks: list of the input genbank filesx
            muscle_location: the path to muscle executable
        Returns:
            None
    '''
    if checkpoint < Checkpoint.SINGLETONS_EXTRACTED:
        logging.info("Extracting sequences for alignment...")
        make_fasta_for_alignments(loci, output)
    logging.info("CHECKPOINT: SINGLETONS_EXTRACTED")
    if checkpoint < Checkpoint.SINGLETONS_ALIGNED:
        logging.info("Aligning sequences...")
        do_alignments(output, cpus, muscle_location)
    logging.info("CHECKPOINT: SINGLETONS_ALIGNED")
    if checkpoint < Checkpoint.ALIGNMENTS_COMBINED:
        logging.info("Making combined alingnment...")
        make_combined_alignment(gbks, output)
    logging.info("CHECKPOINT: ALIGNMENTS_COMBINED")



'''
Create an argument parser using argparse
    Functions:
        get_parser() -> ArgumentParser
        get_config_parser(arg_parser) -> ArgumentParser
        get_blast_parser(arg_parser) -> ArgumentParser
        get_phylo_parser(arg_parser) -> ArgumentParser
        get_records_parser(arg_parser) -> ArgumentParser
        get_search_parser(arg_parser) -> ArgumentParser
        get_seed_parser(arg_parser) -> ArgumentParser
        get_io_parser(arg_parser) -> ArgumentParser
        get_exe_parser(arg_parser) -> ArgumentParser
        get_arguments(arg_parser) -> ArgumentParser
        def parse_args() -> ArgumentParser
'''

import argparse
from argparse import RawTextHelpFormatter

import logging
from getphylo.utils.checkpoint import Checkpoint

def get_parser():
    ''''Create a parser object specific to getphylo'''
    parser = argparse.ArgumentParser(
        "getphylo",
        description="""
        getphylo: a python package to produce heuristic phylogenetic trees from genomic data\n
        example usage: getphylo -g 'input/*.gb' -c 4
        """,
        epilog="Written by Dr. Thom Booth, 2022.",
        formatter_class=RawTextHelpFormatter
        )
    return parser

def get_config_parser(arg_parser):
    '''
    Create an argument group for basic config details.
        Arguments:
            arg_parser: the basic argument parser
        Returns:
            arg_parser: the argument parser with arguments added
    '''
    config_parser = arg_parser.add_argument_group(
            'basic configuration', 'basic configuration of getphylo'
            )
    config_parser.add_argument(
        '-c',
        '--cpus',
        default=1,
        type=int,
        help=(
            'The number of cpus to use for parallelisation\n'
            '(default: %(default)s)'
        )
        )
    config_parser.add_argument(
        '-cp',
        '--checkpoint',
        default='START',
        type=str,
        choices=[cp.name for cp in Checkpoint],
        help=(
            'string indicating the checkpoint to start from:\n'
            'START = default\n '
            'FASTA_EXTRACTED = Skip extracting fasta sequences from genbank files\n '
            'DIAMOND_BUILT = Skip building diamond databases\n '
            'SINGLETONS_IDENTIFIED = Skip identifying singletons from the seed genome\n '
            'SINGLETONS_SEARCHED = Skip searching singletons against other genomes\n '
            'SINGLETONS_THRESHOLDED = Skip thresholding of singletons\n '
            'SINGLETONS_EXTRACTED = Skip extract fasta sequences for alignments\n '
            'SINGLETONS_ALIGNED = Skip individual protein alignments\n '
            'ALIGNMENTS_COMBINED = Skip combining alignments\n '
            'TREES_BUILT = Skip building trees\n '
            'DONE = Done\n '
            '(default: %(default)s)'
        )
    )
    config_parser.add_argument(
        '-l',
        '--logging',
        default='ERROR',
        choices=[
            logging.getLevelName(level) for level in [logging.DEBUG, logging.INFO, logging.WARNING]
            ],
        help='set the logging level\n'
        '(default: %(default)s)'
    )
    return arg_parser

def get_blast_parser(arg_parser):
    '''
    Create an argument group for blastp parameters
        Arguments:
            arg_parser: the basic argument parser
        Returns:
            arg_parser: the argument parser with arguments added
    '''
    phylo_parser = arg_parser.add_argument_group(
        'blastp', 'parameters for blastp'
        )
    phylo_parser.add_argument(
    '-id',
    '--identity',
    type=float,
    default=None,
    help=(
        'minimum identity score for blastp\n'
        '(default: %(default)s)'
    )
    )
    phylo_parser.add_argument(
    '-qc',
    '--query-coverage',
    type=float,
    default=None,
    help=(
        'minimum query coverage for blastp\n'
        '(default: %(default)s)'
    )
    )
    phylo_parser.add_argument(
    '-sc',
    '--subject-coverage',
    type=float,
    default=None,
    help=(
        'minimum subject coverage for blastp\n'
        '(default: %(default)s)'
    )
    )
    return arg_parser


def get_phylo_parser(arg_parser):
    '''
    Create an argument group for phylogenetic analysis
        Arguments:
            arg_parser: the basic argument parser
        Returns:
            arg_parser: the argument parser with arguments added
    '''
    phylo_parser = arg_parser.add_argument_group(
        'phylogenetics', 'parameters for building phylogentic trees'
        )
    phylo_parser.add_argument(
    '-b',
    '--build-all',
    action='store_true',
    help=(
        'build phylogenetic trees for all loci, not just concatenated alignment\n'
        'NOTE: not recommended to use in combination with iqtree\n'
        '(default: %(default)s)'
    )
    )
    phylo_parser.add_argument(
        '-m',
        '--method',
        default='fasttree',
        choices=['fasttree', 'iqtree'],
        help=(
            'choose the phylogenetic method\n'
            'NOTE: Using iqtree will test individual gene models\n'
            'but will exponentially increase the run time\n'
            'NOTE: Not recommended to use in combination with --build-all\n'
            '(default: %(default)s)'
        )
    )
    return arg_parser

def get_records_parser(arg_parser):
    '''
    Create an argument group for ignoring malformed record and annotations.
        Arguments:
            arg_parser: the basic argument parser
        Returns:
            arg_parser: the argument parser with arguments added
    '''
    record_parser = arg_parser.add_argument_group(
            'records and annotations - NOT RECOMMENDED', 
            'deal with malformed records and annotations \n' +
            'it is NOT RECOMMENDED to use these options unless you are confident with their usage'
            )
    record_parser.add_argument(
        '-ia',
        '--ignore-bad-annotations',
        action='store_true',
        help='ignore missing annotations - NOT RECOMMENDED\n (default: %(default)s)'
        )
    record_parser.add_argument(
        '-ir',
        '--ignore-bad-records',
        action='store_true',
        help='ignore poorly formated records - NOT RECOMMENDED\n (default: %(default)s)'
        )
    return arg_parser

def get_search_parser(arg_parser):
    '''
    Create an argument group for changing search parameters
        Arguments:
            arg_parser: the basic argument parser
        Returns:
            arg_parser: the argument parser with arguments added
    '''
    search_parser = arg_parser.add_argument_group(
            'search filtering', 
            'alter the parameters of the orthologue search'
            )
    search_parser.add_argument(
        '-f',
        '--find',
        default=-1,
        type=int,
        help=(
            'integer indicating the number of loci to find in the seed genome\n'
            '(default: %(default)s)'
        )
        )
    search_parser.add_argument(
        '-max',
        '--maxlength',
        default=2000,
        type=int,
        help=(
            'integer indicating the minimum length of loci to be included in the analysis\n'
            '(default: %(default)s)'
        )
        )
    search_parser.add_argument(
        '-min',
        '--minlength',
        default=200,
        type=int,
        help=(
            'integer indicating the minimum length of loci to be included in the analysis\n'
            '(default: %(default)s)'
        )
        )
    search_parser.add_argument(
        '-minl',
        '--minloci',
        default=1,
        type=int,
        help=(
            'minimum number of loci required to continue to alignment and tree building steps\n'
            '(default: %(default)s)'
        )
        )
    search_parser.add_argument(
        '-maxl',
        '--maxloci',
        default=1000,
        type=int,
        help=(
            'maximum number of loci required to continue to alignment and tree building steps\n'
            '(default: %(default)s)'
        )
        )
    search_parser.add_argument(
        '-p',
        '--presence',
        default=100,
        type=float,
        help=(
            'the percentage of genomes each loci must be present in\n'
            '(default: %(default)s)'
        )
        )
    return arg_parser

def get_seed_parser(arg_parser):
    '''
    Create an argument group for changing search parameters
        Arguments:
            arg_parser: the basic argument parser
        Returns:
            arg_parser: the argument parser with arguments added
    '''
    seed_parser = arg_parser.add_argument_group(
            'seed parameters', 
            'seed options for the genome or loci seed'
            )
    seed_parser.add_argument(
        '-r',
        '--random-seed-number',
        default=None,
        type=int,
        help=(
            'integer to be used as a seed for randomising loci selection\n'
            'random if left as None\n'
            '(default: None)'
        )
        )
    seed_parser.add_argument(
        '-s',
        '--seed',
        default=None,
        type=str,
        help='path to a genbankfile with for the target organism\n'
        'first in glob if left as None\n'
        'NOTE: using the smallest genome will generally result in lower runtimes\n'
        'NOTE: this will only effect the results if -p is used\n'
        '(default: %(default)s)'
        )
    return arg_parser

def get_io_parser(arg_parser):
    '''
    Create an argument group for changing input and output parameters
        Arguments:
            arg_parser: the basic argument parser
        Returns:
            arg_parser: the argument parser with arguments added
    '''
    io_parser = arg_parser.add_argument_group(
            'basic input and output', 
            'set options for input and output'
            )
    io_parser.add_argument(
        '-g',
        '--gbks',
        default="*.gbk",
        type=str,
        help='string indicating the genbank files to use in the phylogeny\n'
        '(default: %(default)s)'
        )
    io_parser.add_argument(
        '-o',
        '--output',
        default='output',
        type=str,
        help=(
            'a string designating the name of the folder to output the results\n'
            '(default: %(default)s)'
        )
    )
    io_parser.add_argument(
        '-t',
        '--tag',
        default="locus_tag",
        type=str,
        help='string indicating the GenBank annotations to extract\n'
        '(default: %(default)s)'
        )
    return arg_parser

def get_exe_parser(arg_parser):
    '''
    Create an argument group for providing custom executable paths
        Arguments:
            arg_parser: the basic argument parser
        Returns:
            arg_parser: the argument parser with arguments added
    '''
    exe_parser = arg_parser.add_argument_group(
            'executables',
            'define custom executable paths; the full path or the command if set in PATH\n', 
            )
    exe_parser.add_argument(
        '-d',
        '--diamond',
        default="diamond",
        type=str,
        help=(
        'path to DIAMOND executable \n'
        '(default: %(default)s)'
        )
    )
    exe_parser.add_argument(
        '-mu',
        '--muscle',
        default="muscle",
        type=str,
        help=(
        'path to MUSCLE executable \n'
        '(default: %(default)s)'
        )
    )
    exe_parser.add_argument(
        '-ft',
        '--fasttree',
        default="fasttree",
        type=str,
        help=(
        'path to fasttree2 executable \n'
        '(default: %(default)s)'
        )
    )
    exe_parser.add_argument(
        '-iq',
        '--iqtree',
        default="iqtree",
        type=str,
        help=(
        'path to IQ-TREE executable \n'
        '(default: %(default)s)'
        )
    )
    return arg_parser

def get_arguments(arg_parser):
    '''
    Add arguments and argument groups to the parser
        Arguments:
            arg_parser: the basic argument parser
        Returns:
            arg_parser: the argument parser with arguments added
    '''
    arg_parser = get_io_parser(arg_parser)
    arg_parser = get_config_parser(arg_parser)
    arg_parser = get_search_parser(arg_parser)
    arg_parser = get_blast_parser(arg_parser)
    arg_parser = get_phylo_parser(arg_parser)
    arg_parser = get_seed_parser(arg_parser)
    arg_parser = get_records_parser(arg_parser)
    arg_parser = get_exe_parser(arg_parser)
    return arg_parser

def parse_args():
    '''
    get the arguments from the console via the parser
    '''
    arg_parser = get_parser()
    arg_parser = get_arguments(arg_parser)
    args = arg_parser.parse_args()
    return args



'''
Unique errors for getphylo.
'''

class GetphyloError(Exception):
    '''General class of errors unique to getphylo'''
    pass

class BadInputError(GetphyloError):
    '''Called when user provides bad input'''
    pass
    
class BadSeedError(GetphyloError):
    '''Called when a seed cannot be correctly set'''
    pass

class NoFinalLociError(GetphyloError):
    '''Called when final_loci is empty and cannot be read from final_loci.txt'''
    pass

class NoCandidateLociError(GetphyloError):
    '''Called when candidate_loci is empty and cannot be read from final_loci.txt'''
    pass

class BadAnnotationError(GetphyloError):
    '''
        Called when a genbank files is poorly annotated
        (e.g. duplicate locus tags or missing annotations)
    '''
    pass

class BadRecordError(GetphyloError):
    '''Called when BioPython cannot read records due to misformatting'''
    pass

class FolderExistsError(GetphyloError):
    '''Called by getphylo.utils.io.make_folder when a folder exists.'''
    pass

class FileAlreadyExistsError(GetphyloError):
    '''Called by getphylo.screen when a attempting to write a file and that file already exists.'''
    pass

class InsufficientLociError(GetphyloError):
    '''Called in screen if the number of loci are below the threshold defined by the user'''
    pass

class BadLocusError(GetphyloError):
    '''Called in align when a locus is not present.'''
    pass

class BadExecutableError(GetphyloError):
    '''Called when a non-existant executable path is provided'''
    pass



'''
Runs MUSCLE on a provided fasta file.

Functions:
    run_muscle(filename, outname=None) -> None
    get_muscle_version() -> float
'''
import re
import subprocess
import logging
from getphylo.utils import io

def get_muscle_version(muscle_location: str = 'muscle') -> float:
    '''
    get the muscle version from the command line
        arguments:
            None
        returns
            version_number:
                a float reprisenting the first two parts of the MUSCLE version number
    '''
    with subprocess.Popen(
        [muscle_location, "-version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        ) as process:
        out, _ = process.communicate()
    # only the first line matters
    version = out.decode().splitlines()[0]
    # the second chunk is all that's relevant
    # e.g. "MUSCLE v3.8.1551" ... or muscle "5.1.linux64 ..."
    version = version.split()[1].lower()
    # remove the leading 'v' if present
    version = version.lstrip("v")
    # grab the first bit to floatify
    try:
        version_number = float(re.search(r"\d\.\d+", version)[0])
        return version_number
    except TypeError as error:
        raise RuntimeError("cannot determine version of MUSCLE") from error


def run_muscle(filename: str, outname=None, muscle_location: str = 'muscle') -> None:
    '''
    Run MUSCLE aligner on protein fasta file.
        Arguments:
            filename: path to unaligned sequences
            outname: path for the alignment
        Returns:
            None
    '''
    if outname is None:
        outname = "aligned_" + filename
    # change the argument format depending on the version of MUSCLE
    # also, MUSCLE 5 is much slower than previous versions so print a warning!
    if get_muscle_version(muscle_location) >= 5.0:
        logging.warning(
            'You are using a MUSCLE version 5 or later. '
            'Be aware that MUSCLE 5 is much slower than previous versions.'
        )
        command = [
            muscle_location,
            "-align", filename, 
            "-output", outname
        ]
    else:
        command = [
            muscle_location,
            "-in", filename, 
            "-out", outname
        ]
    io.run_in_command_line(command)



'''
Build trees from a directory containing .fasta alignments

Functions:
    build_all_trees(files: List, cpus: int, method: str, output: str) -> None
    make_trees(output: str, build_all: bool, method: str, cpus: int) -> None
'''
import os
import glob
import logging
from typing import List

from getphylo.utils import io
from getphylo.ext import fasttree, iqtree
from getphylo.utils.errors import GetphyloError

def build_all_trees(
    files: List, cpus: int, method: str, tree_directory: str, output: str, tree_builder:str
    ) -> None:
    '''
    builds all trees in from a list of files
        Arguments:
            files: list of alignment files to be processed
            cpus: number of cpus for parallelisation
            method: pyhlogenetic method (e.g. fasttree)
        Returns:
            None 
    '''
    args_list = []
    if method == 'fasttree':
        for filename in files:
            outfile = os.path.join(
                tree_directory, os.path.basename(io.change_extension(filename, "tree"))
                )
            args_list.append([filename, outfile, tree_builder])
        io.run_in_parallel(fasttree.run_fasttree, args_list, cpus)
    elif method == 'iqtree':
        partition = os.path.join(output, 'partition.txt')
        for filename in files:
            outfile = os.path.join(
                tree_directory, os.path.basename(os.path.splitext(filename)[0])
                )
            args_list.append([filename, outfile, partition, tree_builder])
        io.run_in_parallel(iqtree.run_iqtree, args_list, cpus)
    else:
        raise GetphyloError(method + ' is not a phylogenetic tool.')

def make_trees(output: str, build_all: bool, method: str, cpus: int, tree_builder: str) -> None:
    '''Main routine for trees.
        Arguments:
            output: path to the output directory
        Returns:
            None
    '''
    tree_directory = os.path.join(output, 'trees')
    io.make_folder(tree_directory)
    logging.info("Building trees...")
    if build_all is True:
        files = glob.glob(os.path.join(output, 'aligned_fasta/*.fasta'))
        build_all_trees(files, cpus, method, tree_directory, output, tree_builder)
    else:
        filename = os.path.join(output, 'aligned_fasta/combined_alignment.fasta')
        if method == 'fasttree':
            output = os.path.join(tree_directory, 'combined_alignment.tree')
            fasttree.run_fasttree(filename, output, tree_builder)
        elif method == 'iqtree':
            partition = os.path.join(output, 'partition.txt')
            output = os.path.join(tree_directory, 'combined_alignment')
            iqtree.run_iqtree(filename, output, partition, tree_builder)
        else:
            raise GetphyloError(method + ' is not a phylogenetic tool.')
    logging.info("CHECKPOINT: TREES_BUILT")



'''
Run iqtree.

Functions:
    run_iqtree(alignment_path: str, partition_path: str, out_path: str) -> None

'''
from getphylo.utils import io

def run_iqtree(
    alignment_path: str, out_path: str, partition_path: str=None, iqtree_location: str='iqtree'
    ) -> None:
    '''
    Run fasttree on a protein alignment.
        Arguments:
            alignment_path: path to the alignment
            partition_path: path to the partition file
            out_path: path to the output file
        Returns:
            None
    '''
    command = [
            iqtree_location,
            '-s', alignment_path,
            '-pre', out_path, 
            '-m', 'MFP',
            '-bb', '1000'
            ]
    if partition_path is not None:
        command.append('-spp')
        command.append(partition_path)
    io.run_in_command_line(command)
